import os
import os.path as osp
import subprocess
import sys
import uuid
from concurrent.futures import ThreadPoolExecutor
from enum import Enum
from importlib.util import find_spec
from io import BytesIO
from logging import getLogger
from pathlib import Path, PurePosixPath
from types import ModuleType
from typing import (
    Dict,
    Iterable,
    Iterator,
    List,
    Literal,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    Union,
    cast,
)

import fsspec
from typeguard import typechecked

from .environ import environ
from .registry import Registry

logger = getLogger(__name__)

PathType = Union[str, Path]


class PutFile(Protocol):
    def __call__(
        self,
        local_path: PathType,
        remote_path: PathType,
        mlflow_instance: Optional[ModuleType] = None,
    ) -> str: ...


PUT_FILE_REGISTRY: Registry[PutFile] = Registry(name="put_file")


class FSType(Enum):
    MLFLOW = "mlflow"
    FSSPEC = "fsspec"


class LuxonisFileSystem:
    @typechecked
    def __init__(
        self,
        path: str,
        allow_active_mlflow_run: Optional[bool] = False,
        allow_local: Optional[bool] = True,
        cache_storage: Optional[str] = None,
        put_file_plugin: Optional[str] = None,
    ):
        """Abstraction over remote and local sources.

        Helper class which abstracts uploading and downloading files
        from remote and local sources. Supports S3, MLflow, GCS, and
        local file systems.

        @type path: str
        @param path: Input path consisting of protocol and actual path
            or just path for local files
        @type allow_active_mlflow_run: Optional[bool]
        @param allow_active_mlflow_run: Flag if operations are allowed
            on active MLFlow run. Defaults to False.
        @type allow_local: Optional[bool]
        @param allow_local: Flag if operations are allowed on local file
            system. Defaults to True.
        @type cache_storage: Optional[str]
        @param cache_storage: Path to cache storage. No cache is used if
            set to None. Defaults to None.
        @type put_file_plugin: Optional[str]
        @param put_file_plugin: The name of a registered function under
            the PUT_FILE_REGISTRY to override C{self.put_file}.
        """
        self.cache_storage = cache_storage

        self.protocol, _path = _get_protocol_and_path(path)
        supported_protocols = ["s3", "gcs", "file", "mlflow"]
        if self.protocol not in supported_protocols:
            raise ValueError(
                f"Protocol `{self.protocol}` not supported. Choose from {supported_protocols}."
            )

        _check_package_installed(self.protocol)

        self.allow_local = allow_local
        if self.protocol == "file" and not self.allow_local:
            raise ValueError("Local filesystem is not allowed.")

        if self.protocol == "mlflow":
            self.fs_type = FSType.MLFLOW

            self.allow_active_mlflow_run = allow_active_mlflow_run
            self.is_mlflow_active_run = False
            if _path is not None:
                (
                    self.experiment_id,
                    self.run_id,
                    self.artifact_path,
                ) = self._split_mlflow_path(_path)
            elif _path is None and self.allow_active_mlflow_run:
                self.is_mlflow_active_run = True
                _path = ""
            else:
                raise ValueError(
                    "Using active MLFlow run is not allowed. Specify full MLFlow path."
                )
            self.tracking_uri = environ.MLFLOW_TRACKING_URI

            if self.tracking_uri is None:
                raise KeyError(
                    "There is no 'MLFLOW_TRACKING_URI' in environment variables"
                )
        else:
            self.fs_type = FSType.FSSPEC
            self.fs = self.init_fsspec_filesystem()
        self.path = PurePosixPath(cast(str, _path))

        if put_file_plugin:
            self.put_file = PUT_FILE_REGISTRY.get(put_file_plugin)

    @property
    def is_mlflow(self) -> bool:
        """Returns True if the filesystem is MLFlow.

        @type: bool
        """
        return self.fs_type == FSType.MLFLOW

    @property
    def is_fsspec(self) -> bool:
        """Returns True if the filesystem is fsspec.

        @type: bool
        """
        return self.fs_type == FSType.FSSPEC

    @property
    def full_path(self) -> str:
        """Returns full remote path.

        @type: str
        """
        return f"{self.protocol}://{self.path}"

    def init_fsspec_filesystem(self) -> fsspec.AbstractFileSystem:
        """Initializes L{fsspec} filesystem based on the used protocol.

        @rtype: L{fsspec.AbstractFileSystem}
        @return: Initialized fsspec filesystem.
        """
        if self.protocol == "s3":
            # NOTE: In theory boto3 should look in environment variables automatically but it doesn't seem to work
            fs = fsspec.filesystem(
                self.protocol,
                key=environ.AWS_ACCESS_KEY_ID,
                secret=environ.AWS_SECRET_ACCESS_KEY,
                endpoint_url=environ.AWS_S3_ENDPOINT_URL,
            )
        elif self.protocol == "gcs":
            if environ.GOOGLE_APPLICATION_CREDENTIALS is None:
                raise KeyError(
                    "There is no 'GOOGLE_APPLICATION_CREDENTIALS' in environment variables"
                )
            # NOTE: This should automatically read from GOOGLE_APPLICATION_CREDENTIALS
            fs = fsspec.filesystem(self.protocol)
        elif self.protocol == "file":
            fs = fsspec.filesystem(self.protocol)
        else:
            raise NotImplementedError
        if self.cache_storage is None:
            return fs

        if self.protocol == "file":
            logger.warning("Ignoring cache storage for local filesystem.")
            return fs

        return fsspec.filesystem(
            "filecache", fs=fs, cache_storage=self.cache_storage
        )

    def put_file(
        self,
        local_path: PathType,
        remote_path: PathType,
        mlflow_instance: Optional[ModuleType] = None,
    ) -> str:
        """Copy a single file to remote storage.

        @type local_path: PathType
        @param local_path: Path to local file
        @type remote_path: PathType
        @param remote_path: Relative path to remote file
        @type mlflow_instance: Optional[L{ModuleType}]
        @param mlflow_instance: MLFlow instance if uploading to active
            run. Defaults to C{None}.
        @rtype: str
        @return: The full remote path of the uploded file.
        """
        local_path = str(local_path)
        if self.is_mlflow:
            # NOTE: remote_path not used in mlflow since it creates new folder each time
            if self.is_mlflow_active_run:
                if mlflow_instance is not None:
                    mlflow_instance.log_artifact(local_path)
                else:
                    raise KeyError("No active mlflow_instance provided.")
            else:
                import mlflow

                client = mlflow.MlflowClient(tracking_uri=self.tracking_uri)
                client.log_artifact(run_id=self.run_id, local_path=local_path)

        elif self.is_fsspec:
            self.fs.put_file(local_path, str(self.path / remote_path))
        return self.protocol + "://" + str(self.path / remote_path)

    def put_dir(
        self,
        local_paths: Union[PathType, Iterable[PathType]],
        remote_dir: PathType,
        uuid_dict: Optional[Dict[str, str]] = None,
        mlflow_instance: Optional[ModuleType] = None,
    ) -> Optional[Dict[str, str]]:
        """Uploads files to remote storage.

        @type local_paths: Union[PathType, Sequence[PathType]]
        @param local_paths: Either a string specifying a directory to
            walk the files or a list of files which can be in different
            directories.
        @type remote_dir: PathType
        @param remote_dir: Relative path to remote directory
        @type uuid_dict: Optional[Dict[str, str]]
        @param uuid_dict: Stores paths as keys and corresponding UUIDs
            as values to replace the file basename.
        @type mlflow_instance: Optional[L{ModuleType}]
        @param mlflow_instance: MLFlow instance if uploading to active
            run. Defaults to None.
        @rtype: Optional[Dict[str, str]]
        @return: When local_paths is a list, this maps local_paths to
            remote_paths
        """
        if self.is_mlflow:
            raise NotImplementedError
        elif self.is_fsspec:
            if isinstance(local_paths, (Path, str)):
                local_paths = Path(local_paths)
                if not Path(local_paths).is_dir():
                    raise ValueError("Path must be a directory.")
                self.fs.put(
                    str(local_paths),
                    str(self.path / remote_dir),
                    recursive=True,
                )
            else:
                upload_dict = {}
                with ThreadPoolExecutor() as executor:
                    for local_path in local_paths:
                        local_path = str(local_path)
                        if uuid_dict is not None:
                            file_uuid = uuid_dict[local_path]
                            ext = osp.splitext(local_path)[1]
                            basename = file_uuid + ext
                        else:
                            basename = Path(local_path).name
                        remote_path = str(PurePosixPath(remote_dir) / basename)
                        upload_dict[local_path] = remote_path
                        executor.submit(self.put_file, local_path, remote_path)
                return upload_dict

    def put_bytes(
        self,
        file_bytes: bytes,
        remote_path: PathType,
        mlflow_instance: Optional[ModuleType] = None,
    ) -> None:
        """Uploads a file to remote storage directly from file bytes.

        @type file_bytes: bytes
        @param file_bytes: the bytes for the file contents
        @type remote_path: PathType
        @param remote_path: Relative path to remote file
        @type mlflow_instance: Optional[L{ModuleType}]
        @param mlflow_instance: MLFlow instance if uploading to active
            run. Defaults to None.
        """
        if self.is_mlflow:
            raise NotImplementedError
        elif self.is_fsspec:
            full_path = str(self.path / remote_path)
            with self.fs.open(full_path, "wb") as file:
                file.write(file_bytes)  # type: ignore

    def get_file(
        self,
        remote_path: PathType,
        local_path: PathType,
        mlflow_instance: Optional[ModuleType] = None,
    ) -> Path:
        """Copy a single file from remote storage.

        @type remote_path: PathType
        @param remote_path: Relative path to remote file
        @type local_path: PathType
        @param local_path: Path to local file
        @type mlflow_instance: Optional[L{ModuleType}]
        @param mlflow_instance: MLFlow instance if uploading to active
            run. Defaults to C{None}.
        @rtype: Path
        @return: Path to the downloaded file.
        """

        if self.is_mlflow:
            raise NotImplementedError
        elif self.is_fsspec:
            self.fs.download(
                str(self.path / remote_path),
                str(local_path),
                recursive=False,
            )
        return Path(local_path) / Path(remote_path).name

    def delete_file(self, remote_path: PathType) -> None:
        """Deletes a single file from remote storage.

        @type remote_path: PathType
        @param remote_path: Relative path to remote file
        """
        if self.is_fsspec:
            full_remote_path = str(self.path / remote_path)
            self.fs.rm(full_remote_path)
        else:
            raise NotImplementedError

    def delete_files(self, remote_paths: List[PathType]) -> None:
        """Deletes multiple files from remote storage.

        @type remote_paths: List[PathType]
        @param remote_paths: Relative paths to remote files
        """
        if self.is_fsspec:
            full_remote_paths = [
                str(self.path / remote_path) for remote_path in remote_paths
            ]
            self.fs.rm(full_remote_paths)
        else:
            raise NotImplementedError

    def get_dir(
        self,
        remote_paths: Union[PathType, Sequence[PathType]],
        local_dir: PathType,
        mlflow_instance: Optional[ModuleType] = None,
    ) -> Path:
        """Copies many files from remote storage to local storage.

        @type remote_paths: Union[PathType, Sequence[PathType]]
        @param remote_paths: Either a string specifying a directory to
            walk the files or a list of files which can be in different
            directories.
        @type local_dir: PathType
        @param local_dir: Path to local directory
        @type mlflow_instance: Optional[L{ModuleType}]
        @param mlflow_instance: MLFlow instance if uploading to active
            run. Defaults to C{None}.
        @rtype: Path
        @return: Path to the downloaded directory.
        """
        if self.is_mlflow:
            raise NotImplementedError
        elif self.is_fsspec:
            if isinstance(remote_paths, Path) or isinstance(remote_paths, str):
                self.fs.download(
                    str(self.path / remote_paths),
                    str(local_dir),
                    recursive=True,
                )
                return Path(local_dir) / Path(remote_paths).name

            elif isinstance(remote_paths, list):
                with ThreadPoolExecutor() as executor:
                    for remote_path in remote_paths:
                        local_path = str(
                            local_dir / Path(Path(remote_path).name)
                        )
                        executor.submit(self.get_file, remote_path, local_path)

        return Path(local_dir)

    def delete_dir(
        self, remote_dir: PathType = "", allow_delete_parent: bool = False
    ) -> None:
        """Deletes a directory and all its contents from remote storage.

        @type remote_dir: PathType
        @param remote_dir: Relative path to remote directory.
        @type allow_delete_parent: bool
        @param allow_delete_parent: If True, allows deletion of the
            parent directory.
        """
        if not remote_dir and not allow_delete_parent:
            raise ValueError(
                f"No directory specified, this would the parent directory at `{self.path}`."
                "If this is your intention, you must pass `allow_delete_parent=True`."
            )

        elif remote_dir is None:
            remote_dir = ""

        if self.is_fsspec:
            full_remote_dir = str(self.path / remote_dir)
            self.fs.rm(full_remote_dir, recursive=True)
        else:
            raise NotImplementedError

    def walk_dir(
        self,
        remote_dir: PathType,
        recursive: bool = True,
        typ: Literal["file", "directory", "all"] = "file",
    ) -> Iterator[str]:
        """Walks through the individual files in a remote directory.

        @type remote_dir: PathType
        @param remote_dir: Relative path to remote directory
        @type recursive: bool
        @param recursive: If True, walks through the directory
            recursively.
        @type typ: Literal["file", "directory", "all"]
        @param typ: Specifies the type of files to walk through.
            Defaults to "file".
        @rtype: Iterator[str]
        @return: Iterator over the paths.
        """

        if self.is_mlflow:
            raise NotImplementedError
        elif self.is_fsspec:
            full_path = str(self.path / remote_dir)
            for file in self.fs.ls(full_path, detail=True):
                name = str(
                    PurePosixPath(str(file["name"])).relative_to(self.path)
                )
                if typ == "all" or file["type"] == typ:
                    yield name
                if recursive and file["type"] == "directory":
                    yield from self.walk_dir(name, recursive, typ)

    def read_to_byte_buffer(
        self, remote_path: Optional[PathType] = None
    ) -> BytesIO:
        """Reads a file into a byte buffer.

        @type remote_path: Optional[PathType]
        @param remote_path: Relative path to remote file.
        @rtype: BytesIO
        @return: The byte buffer containing the file contents.
        """

        if self.is_mlflow:
            if self.is_mlflow_active_run:
                raise ValueError(
                    "Reading to byte buffer not available for active mlflow runs."
                )
            else:
                if self.artifact_path is None:
                    raise ValueError("No relative artifact path specified.")
                import mlflow

                client = mlflow.MlflowClient(tracking_uri=self.tracking_uri)
                if self.run_id is None:
                    raise RuntimeError(
                        "`run_id` cannot be `None` when using `mlflow`"
                    )
                download_path = client.download_artifacts(
                    run_id=self.run_id, path=self.artifact_path, dst_path="."
                )
            with open(download_path, "rb") as f:
                buffer = BytesIO(f.read())
            os.remove(download_path)  # remove local file

        else:
            if remote_path is not None:
                download_path = str(self.path / remote_path)
            else:
                download_path = str(self.path)
            with self.fs.open(download_path, "rb") as f:
                buffer = BytesIO(cast(bytes, f.read()))

        return buffer

    def get_file_uuid(self, path: PathType, local: bool = False) -> str:
        """Reads a file and returns the (unique) UUID generated from
        file bytes.

        @type path: PathType
        @param path: Relative path to remote file.
        @type local: bool
        @param local: Specifies a local path as opposed to a remote
            path.
        @rtype: str
        @return: The generated UUID.
        """

        if local:
            with open(path, "rb") as f:
                file_contents = cast(bytes, f.read())
        elif self.is_mlflow:
            raise NotImplementedError
        else:
            download_path = str(self.path / path)
            with self.fs.open(download_path, "rb") as f:
                file_contents = cast(bytes, f.read())

        file_hash_uuid = str(
            uuid.uuid5(uuid.NAMESPACE_URL, file_contents.hex())
        )

        return file_hash_uuid

    def get_file_uuids(
        self, paths: Iterable[PathType], local: bool = False
    ) -> Dict[str, str]:
        """Computes the UUIDs for all files stored in the filesystem.

        @type paths: List[PathType]
        @param paths: A list of relative remote paths if remote else
            local paths.
        @type local: bool
        @param local: Specifies local paths as opposed to remote paths.
        @rtype: Dict[str, str]
        @return: A dictionary mapping the paths to their UUIDs
        """

        result = {}

        if self.is_fsspec:
            with ThreadPoolExecutor() as executor:
                for path in paths:
                    path = str(path)
                    future = executor.submit(self.get_file_uuid, path, local)
                    result[path] = future.result()

        return result

    @staticmethod
    def _split_mlflow_path(path: PathType) -> List[Optional[str]]:
        """Splits mlflow path into 3 parts."""
        path = Path(path)
        parts: List[Optional[str]] = list(path.parts)
        if len(parts) < 3:
            while len(parts) < 3:
                parts.append(None)
        elif len(parts) > 3:
            parts[2] = "/".join(cast(List[str], parts[2:]))
            parts = parts[:3]
        return parts

    def is_directory(self, remote_path: PathType) -> bool:
        """Checks whether the given remote path is a directory.

        @type remote_path: PathType
        @param remote_path: Relative path to remote file.
        @rtype: bool
        @return: True if the path is a directory.
        """

        full_path = str(self.path / remote_path)
        file_info = self.fs.info(full_path)
        return file_info["type"] == "directory"

    def exists(self, remote_path: PathType = "") -> bool:
        """Checks whether the given remote path exists.

        @type remote_path: PathType
        @param remote_path: Relative path to remote file. Defaults to ""
            (root).
        @rtype: bool
        @return: True if the path exists.
        """
        full_path = str(self.path / remote_path)
        return self.fs.exists(full_path)

    @staticmethod
    def split_full_path(path: PathType) -> Tuple[str, str]:
        """Splits the full path into protocol and absolute path.

        @type path: PathType
        @param path: Full path
        @rtype: Tuple[str, str]
        @return: Tuple of protocol and absolute path.
        """
        path = str(path).rstrip("/\\")
        return osp.split(path)

    @staticmethod
    def get_protocol(path: str) -> str:
        """Extracts the detected protocol from a path.

        @type path: str
        @param path: Full path
        @rtype: str
        @return: Protocol of the path.
        """
        return _get_protocol_and_path(path)[0]

    @staticmethod
    def download(url: str, dest: Optional[PathType]) -> Path:
        """Downloads file or directory from remote storage.

        Intended for downloading a single remote object, elevating the
        need to create an instance of L{LuxonisFileSystem}.

        @type url: str
        @param url: URL to the file or directory
        @type dest: Optional[PathType]
        @param dest: Destination directory. If unspecified, the current
            directory is used.
        @rtype: Path
        @return: Path to the downloaded file or directory.
        """

        if LuxonisFileSystem.get_protocol(url) == "file":
            return Path(url)

        dest = Path(dest or ".")
        absolute_path, remote_path = LuxonisFileSystem.split_full_path(url)
        if dest.suffix:
            local_path = dest
        else:
            local_path = dest / remote_path
        fs = LuxonisFileSystem(absolute_path)

        if fs.is_directory(remote_path):
            fs.get_dir(remote_path, local_path)
        else:
            fs.get_file(remote_path, str(local_path))

        return local_path

    @staticmethod
    def upload(local_path: PathType, url: str) -> None:
        """Uploads file or directory to remote storage.

        Intended for uploading a single local object, elevating the need
        to create an instance of L{LuxonisFileSystem}.

        @type local_path: PathType
        @param local_path: Path to the local file or directory
        @type url: str
        @param url: URL to the remote file or directory
        @rtype: str
        @return: URL to the uploaded file or directory.
        """

        absolute_path, remote_path = LuxonisFileSystem.split_full_path(url)
        fs = LuxonisFileSystem(absolute_path)
        if Path(local_path).is_dir():
            fs.put_dir(local_path, remote_path)
        else:
            fs.put_file(str(local_path), remote_path)


def _check_package_installed(protocol: str) -> None:  # pragma: no cover
    if protocol in ["gs", "gcs"] and find_spec("gcsfs") is None:
        _pip_install(protocol, "gcsfs", "2023.3.0")
    elif protocol == "s3" and find_spec("s3fs") is None:
        _pip_install(protocol, "s3fs", "2023.3.0")
    elif protocol == "mlflow" and find_spec("mlflow") is None:
        _pip_install(protocol, "mlflow", "2.10.0")


def _get_protocol_and_path(path: str) -> Tuple[str, Optional[str]]:
    """Gets the protocol and absolute path of a full path."""

    if "://" in path:
        protocol, path = path.split("://")
        if protocol == "gs":
            # ensure gs:// URLs are accepted as the gcs protocol
            protocol = "gcs"
    else:
        # assume that it is local path
        protocol = "file"

    return protocol, path if path else None


def _pip_install(protocol: str, package: str, version: str) -> None:
    logger.error(f"'{package}' is necessary for '{protocol}://' protocol.")
    logger.info(f"Installing {package}...")
    subprocess.run(
        [sys.executable, "-m", "pip", "install", f"{package}>={version}"]
    )



import shutil
import tempfile
from pathlib import Path
from typing import List, Optional

import pytest

from luxonis_ml.utils import environ
from luxonis_ml.utils.filesystem import (
    LuxonisFileSystem,
    _get_protocol_and_path,
)

URL_PATH = "luxonis-test-bucket/luxonis-ml-test-data/fs_test_data"

LOCAL_ROOT = Path("tests/data/test_filesystem")
LOCAL_FILE_PATH = LOCAL_ROOT / "_file.txt"
LOCAL_DIR_PATH = LOCAL_ROOT / "_dir"

skip_if_no_s3_credentials = pytest.mark.skipif(
    environ.AWS_ACCESS_KEY_ID is None
    or environ.AWS_SECRET_ACCESS_KEY is None
    or environ.AWS_S3_ENDPOINT_URL is None,
    reason="S3 credentials not set",
)

skip_if_no_gcs_credentials = pytest.mark.skipif(
    environ.GOOGLE_APPLICATION_CREDENTIALS is None,
    reason="GCS credentials not set",
)


# NOTE: needed for tests running in GitHub Actions using the matrix strategy
#       to avoid race conditions when running tests in parallel
def get_os_python_specific_url(
    protocol: str, platform: str, python_version: str
):
    return f"{protocol}://{URL_PATH}_{platform}_{python_version}"


@pytest.fixture
def fs(request, python_version: str, platform_name: str):
    url_path = get_os_python_specific_url(
        request.param, platform_name, python_version
    )
    yield LuxonisFileSystem(url_path)


def parametrize_dependent_fixture(
    name: str = "fs",
    depends: Optional[List[str]] = None,
):
    depends = depends or []
    skips = {
        "gs": skip_if_no_gcs_credentials,
        "s3": skip_if_no_s3_credentials,
    }

    def decorator(func):
        protocols = ["gs", "s3"]
        return pytest.mark.parametrize(
            name,
            [
                pytest.param(
                    protocol,
                    marks=[
                        pytest.mark.dependency(
                            depends=[f"{d}_{protocol}" for d in depends],
                            name=f"{func.__name__}_{protocol}",
                        ),
                        skips[protocol],
                    ],
                )
                for protocol in protocols
            ],
            indirect=name == "fs",
        )(func)

    return decorator


def clean(fs: LuxonisFileSystem, name: str):
    if Path(name).suffix:
        fs.delete_file(name)
    else:
        fs.delete_dir(name)
    assert not fs.exists(name)


@pytest.fixture(scope="function", autouse=True)
def setup_remote_tests():
    LOCAL_ROOT.mkdir(parents=True, exist_ok=True)
    LOCAL_DIR_PATH.mkdir(parents=True, exist_ok=True)
    LOCAL_FILE_PATH.write_text("test 42")

    for i in range(5):
        file_path = LOCAL_DIR_PATH / f"file_{i}.txt"
        file_path.write_text(f"file {i + 21}")

    yield

    shutil.rmtree(LOCAL_ROOT)


def test_protocol():
    assert _get_protocol_and_path("foo://bar/baz") == ("foo", "bar/baz")
    assert _get_protocol_and_path("gs://foo/bar") == ("gcs", "foo/bar")
    assert _get_protocol_and_path("local_path/to/file") == (
        "file",
        "local_path/to/file",
    )

    with pytest.raises(ValueError):
        LuxonisFileSystem("foo://bar")


def test_fail():
    with pytest.raises(ValueError):
        LuxonisFileSystem(str(LOCAL_FILE_PATH), allow_local=False)


@parametrize_dependent_fixture()
def test_file_download(fs: LuxonisFileSystem):
    with tempfile.TemporaryDirectory() as tempdir:
        file_path = fs.get_file("file.txt", tempdir)
        assert file_path.exists()
        assert file_path.read_text() == "file\n"


@parametrize_dependent_fixture()
def test_bytes(fs: LuxonisFileSystem):
    fs.put_bytes(b"bytes_test", "test_bytes.txt")
    assert fs.exists("test_bytes.txt")
    buffer = fs.read_to_byte_buffer("test_bytes.txt")
    assert buffer.read() == b"bytes_test"

    clean(fs, "test_bytes.txt")


@parametrize_dependent_fixture(depends=["test_file_download"])
def test_file_upload(fs: LuxonisFileSystem):
    fs.put_file(LOCAL_FILE_PATH, "_file_upload_test.txt")
    assert fs.exists("_file_upload_test.txt")

    with tempfile.TemporaryDirectory() as tempdir:
        file_path = fs.get_file("_file_upload_test.txt", tempdir)
        assert file_path.exists()
        assert file_path.read_text() == LOCAL_FILE_PATH.read_text()

    clean(fs, "_file_upload_test.txt")


@parametrize_dependent_fixture()
def test_dir_download(fs: LuxonisFileSystem):
    with tempfile.TemporaryDirectory() as tempdir:
        dir_path = fs.get_dir("dir", tempdir)

        assert dir_path.exists()
        for i in range(1, 6):
            file_path = Path(dir_path, f"file_{i}.txt")
            assert file_path.exists()
            assert file_path.read_text() == f"file_{i}\n"

    with tempfile.TemporaryDirectory() as tempdir:
        dir_path = fs.get_dir(
            [
                "dir/file_1.txt",
                "dir/file_2.txt",
                "dir/file_3.txt",
                "dir/file_4.txt",
                "dir/file_5.txt",
            ],
            tempdir,
        )

        assert dir_path.exists()
        for i in range(1, 6):
            file_path = Path(dir_path, f"file_{i}.txt")
            assert file_path.exists()
            assert file_path.read_text() == f"file_{i}\n"


@parametrize_dependent_fixture(depends=["test_dir_download"])
def test_dir_upload(fs: LuxonisFileSystem):
    if fs.exists("_dir_upload_test"):  # pragma: no cover
        fs.delete_dir("_dir_upload_test")
    assert not fs.exists("_dir_upload_test")

    fs.put_dir(LOCAL_DIR_PATH, "_dir_upload_test")
    assert fs.exists("_dir_upload_test")
    assert fs.is_directory("_dir_upload_test")

    def check_contents(dir_path: Path):
        assert dir_path.exists()
        for i in range(5):
            file_path = Path(dir_path, f"file_{i}.txt")
            assert file_path.exists()
            assert (
                file_path.read_text()
                == (LOCAL_DIR_PATH / f"file_{i}.txt").read_text()
            )

    with tempfile.TemporaryDirectory() as tempdir:
        dir_path = fs.get_dir("_dir_upload_test", tempdir)
        check_contents(dir_path)

    fs.delete_files([f"_dir_upload_test/file_{i}.txt" for i in range(4)])
    clean(fs, "_dir_upload_test")

    fs.put_dir([str(p) for p in LOCAL_DIR_PATH.iterdir()], "__dir_upload_test")
    assert fs.exists("__dir_upload_test")

    with tempfile.TemporaryDirectory() as tempdir:
        dir_path = fs.get_dir("__dir_upload_test", tempdir)
        check_contents(dir_path)

    clean(fs, "__dir_upload_test")


@parametrize_dependent_fixture()
def test_walk_dir(fs: LuxonisFileSystem):
    walked_files = [Path(file).name for file in fs.walk_dir("dir")]
    assert set(walked_files) == {f"file_{i}.txt" for i in range(1, 6)}


@parametrize_dependent_fixture(
    "protocol",
    [
        "test_file_download",
        "test_dir_download",
    ],
)
def test_static_download(
    protocol: str, python_version: str, platform_name: str
):
    url_root = get_os_python_specific_url(
        protocol, platform_name, python_version
    )
    with tempfile.TemporaryDirectory() as tempdir:
        url = f"{url_root}/file.txt"
        path = LuxonisFileSystem.download(url, tempdir)
        assert path.exists()
        assert path.read_text() == "file\n"

        url = f"{url_root}/dir"
        path = LuxonisFileSystem.download(url, tempdir)
        assert path.exists()
        for i in range(1, 6):
            file_path = Path(path, f"file_{i}.txt")
            assert file_path.exists()
            assert file_path.read_text() == f"file_{i}\n"


@parametrize_dependent_fixture(
    "protocol",
    [
        "test_file_upload",
        "test_dir_upload",
        "test_static_download",
    ],
)
def test_static_upload(protocol: str, python_version: str, platform_name: str):
    url_root = get_os_python_specific_url(
        protocol, platform_name, python_version
    )
    with tempfile.TemporaryDirectory() as tempdir:
        url = f"{url_root}/_file_upload_test.txt"
        LuxonisFileSystem.upload(LOCAL_FILE_PATH, url)

        file_path = LuxonisFileSystem.download(url, tempdir)
        assert file_path.exists()
        assert file_path.read_text() == LOCAL_FILE_PATH.read_text()

        url = f"{url_root}/_dir_upload_test"
        LuxonisFileSystem.upload(LOCAL_DIR_PATH, url)

        dir_path = LuxonisFileSystem.download(url, tempdir)
        assert dir_path.exists()
        for i in range(5):
            file_path = Path(dir_path, f"file_{i}.txt")
            assert file_path.exists()
            assert (
                file_path.read_text()
                == (LOCAL_DIR_PATH / f"file_{i}.txt").read_text()
            )

    fs = LuxonisFileSystem(url_root)
    clean(fs, "_file_upload_test.txt")
    clean(fs, "_dir_upload_test")



from contextlib import suppress
from typing import Any, Dict, List, Optional

from pydantic import Field, model_validator
from typing_extensions import Self

from luxonis_ml.utils import BaseModelExtraForbid

from ...utils import infer_layout
from ..enums import DataType, InputType


class PreprocessingBlock(BaseModelExtraForbid):
    """Represents preprocessing operations applied to the input data.

    @type mean: list | None
    @ivar mean: Mean values in channel order. Order depends on the order
        in which the model was trained on.
    @type scale: list | None
    @ivar scale: Standardization values in channel order. Order depends
        on the order in which the model was trained on.
    @type reverse_channels: bool | None
    @ivar reverse_channels: If True input to the model is RGB else BGR.
    @type interleaved_to_planar: bool | None
    @ivar interleaved_to_planar: If True input to the model is
        interleaved (NHWC) else planar (NCHW).
    @type dai_type: str | None
    @ivar dai_type: DepthAI input type which is read by DepthAI to
        automatically setup the pipeline.
    """

    mean: Optional[List[float]] = Field(
        None,
        description="Mean values in channel order. Order depends on the order in which the model was trained on.",
    )
    scale: Optional[List[float]] = Field(
        None,
        description="Standardization values in channel order. Order depends on the order in which the model was trained on.",
    )
    reverse_channels: Optional[bool] = Field(
        None,
        deprecated="Deprecated, use `dai_type` instead.",
        description="If True input to the model is RGB else BGR.",
    )
    interleaved_to_planar: Optional[bool] = Field(
        None,
        deprecated="Deprecated, use `dai_type` instead.",
        description="If True input to the model is interleaved (NHWC) else planar (NCHW).",
    )
    dai_type: Optional[str] = Field(
        None,
        description="DepthAI input type which is read by DepthAI to automatically setup the pipeline.",
    )


class Input(BaseModelExtraForbid):
    """Represents input stream of a model.

    @type name: str
    @ivar name: Name of the input layer.

    @type dtype: DataType
    @ivar dtype: Data type of the input data (e.g., 'float32').

    @type input_type: InputType
    @ivar input_type: Type of input data (e.g., 'image').

    @type shape: list
    @ivar shape: Shape of the input data as a list of integers (e.g. [H,W], [H,W,C], [N,H,W,C], ...).

    @type layout: str
    @ivar layout: Lettercode interpretation of the input data dimensions (e.g., 'NCHW').

    @type preprocessing: PreprocessingBlock
    @ivar preprocessing: Preprocessing steps applied to the input data.
    """

    name: str = Field(description="Name of the input layer.")
    dtype: DataType = Field(
        description="Data type of the input data (e.g., 'float32')."
    )
    input_type: InputType = Field(
        description="Type of input data (e.g., 'image')."
    )
    shape: List[int] = Field(
        min_length=1,
        description="Shape of the input data as a list of integers (e.g. [H,W], [H,W,C], [N,H,W,C], ...).",
    )
    layout: str = Field(
        "NCHW",
        description="Lettercode interpretation of the input data dimensions (e.g., 'NCHW')",
        min_length=1,
    )
    preprocessing: PreprocessingBlock = Field(
        description="Preprocessing steps applied to the input data."
    )

    @model_validator(mode="after")
    def validate_layout(self) -> Self:
        self.layout = self.layout.upper()

        if len(self.layout) != len(set(self.layout)):
            raise ValueError("Layout must not contain duplicate letters.")

        if len(self.layout) != len(self.shape):
            raise ValueError("Layout and shape must have the same length.")

        if "N" in self.layout and self.layout[0] != "N":
            raise ValueError(
                "If N (batch size) is included in the layout, it must be first"
            )

        if self.input_type == InputType.IMAGE:
            if "C" not in self.layout:
                raise ValueError(
                    "C letter must be present in layout for image input type."
                )

        return self

    @model_validator(mode="before")
    @staticmethod
    def infer_layout(data: Dict[str, Any]) -> Dict[str, Any]:
        if "shape" in data and "layout" not in data:
            with suppress(Exception):
                data["layout"] = infer_layout(data["shape"])
        return data



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Tuple


class VectorDBAPI(ABC):
    """Abstract class for Vector Database APIs.

    This class defines a common interface for vector database operations
    for different implementations like Qdrant and Weaviate.
    """

    @abstractmethod
    def create_collection(
        self, collection_name: str, properties: List[str]
    ) -> None:
        """Create a collection in the vector database."""
        pass

    @abstractmethod
    def delete_collection(self) -> None:
        """Delete the collection from the vector database."""
        pass

    @abstractmethod
    def insert_embeddings(
        self,
        ids: List[str],
        embeddings: List[List[float]],
        payloads: List[Dict[str, Any]],
        batch_size: int = 100,
    ) -> None:
        """Insert embeddings into the collection."""
        pass

    @abstractmethod
    def search_similar_embeddings(
        self, embedding: List[float], top_k: int = 10
    ) -> Tuple[List[str], List[float]]:
        """Search for similar embeddings in the collection."""
        pass

    @abstractmethod
    def get_similarity_scores(
        self,
        reference_id: str,
        other_ids: List[str],
        sort_distances: bool = True,
    ) -> Tuple[List[str], List[float]]:
        """Get similarity scores between a reference embedding and other
        embeddings."""
        pass

    @abstractmethod
    def compute_similarity_matrix(self) -> List[List[float]]:
        """Compute a similarity matrix for all the embeddings in the
        collection."""
        pass

    @abstractmethod
    def retrieve_embeddings_by_ids(self, ids: List[str]) -> List[List[float]]:
        """Retrieve embeddings associated with a list of IDs."""
        pass

    @abstractmethod
    def retrieve_all_embeddings(self) -> Tuple[List[str], List[List[float]]]:
        """Retrieve all embeddings from the collection."""
        pass

    @abstractmethod
    def retrieve_all_ids(self) -> List[str]:
        """Retrieve all IDs from the collection."""
        pass

    @abstractmethod
    def retrieve_payloads_by_ids(
        self, ids: List[str], properties: Optional[List[str]]
    ) -> List[Dict[str, Any]]:
        """Retrieve payloads associated with a list of IDs."""
        pass



import json
import shutil
from collections import defaultdict
from pathlib import Path
from typing import Dict, Final, List, Set

import cv2
import numpy as np
import pytest

from luxonis_ml.data import (
    Augmentations,
    BucketStorage,
    LuxonisDataset,
    LuxonisLoader,
    LuxonisParser,
    LuxonisSource,
)
from luxonis_ml.data.utils.data_utils import rgb_to_bool_masks
from luxonis_ml.enums import DatasetType
from luxonis_ml.utils import LuxonisFileSystem

SKELETONS: Final[dict] = {
    "keypoints": (
        [
            "nose",
            "left_eye",
            "right_eye",
            "left_ear",
            "right_ear",
            "left_shoulder",
            "right_shoulder",
            "left_elbow",
            "right_elbow",
            "left_wrist",
            "right_wrist",
            "left_hip",
            "right_hip",
            "left_knee",
            "right_knee",
            "left_ankle",
            "right_ankle",
        ],
        [
            [15, 13],
            [13, 11],
            [16, 14],
            [14, 12],
            [11, 12],
            [5, 11],
            [6, 12],
            [5, 6],
            [5, 7],
            [6, 8],
            [7, 9],
            [8, 10],
            [1, 2],
            [0, 1],
            [0, 2],
            [1, 3],
            [2, 4],
            [3, 5],
            [4, 6],
        ],
    ),
}
URL_PREFIX: Final[str] = "gs://luxonis-test-bucket/luxonis-ml-test-data"
WORK_DIR: Final[str] = "tests/data/parser_datasets"
DATASET_NAME: Final[str] = "test-dataset"
TASKS: Final[Set[str]] = {
    "segmentation",
    "classification",
    "keypoints",
    "boundingbox",
}
DATA_DIR = Path("tests/data/test_dataset")


@pytest.fixture(autouse=True, scope="module")
def prepare_dir():
    DATA_DIR.mkdir(parents=True, exist_ok=True)

    yield

    shutil.rmtree(DATA_DIR)


def make_image(i) -> Path:
    path = DATA_DIR / f"img_{i}.jpg"
    if not path.exists():
        img = np.zeros((512, 512, 3), dtype=np.uint8)
        img[0:10, 0:10] = np.random.randint(
            0, 255, (10, 10, 3), dtype=np.uint8
        )
        cv2.imwrite(str(path), img)
    return path


@pytest.mark.parametrize(
    ("bucket_storage",),
    [
        (BucketStorage.LOCAL,),
        (BucketStorage.S3,),
        (BucketStorage.GCS,),
    ],
)
def test_dataset(
    bucket_storage: BucketStorage,
    platform_name: str,
    python_version: str,
    subtests,
):
    dataset_name = f"{DATASET_NAME}-{bucket_storage.value}-{platform_name}-{python_version}"
    with subtests.test("test_create", bucket_storage=bucket_storage):
        parser = LuxonisParser(
            f"{URL_PREFIX}/COCO_people_subset.zip",
            dataset_name=dataset_name,
            save_dir=WORK_DIR,
            dataset_type=DatasetType.COCO,
            bucket_storage=bucket_storage,
            delete_existing=True,
            delete_remote=True,
        )
        parser.parse()
        dataset = LuxonisDataset(dataset_name, bucket_storage=bucket_storage)
        assert LuxonisDataset.exists(
            dataset_name, bucket_storage=bucket_storage
        )
        assert dataset.get_classes()[0] == ["person"]
        assert set(dataset.get_tasks()) == TASKS
        assert dataset.get_skeletons() == SKELETONS
        assert dataset.identifier == dataset_name

    if "dataset" not in locals():
        pytest.exit("Dataset creation failed")

    with subtests.test("test_source"):
        print(dataset.source.to_document())
        assert dataset.source.to_document() == LuxonisSource().to_document()
        dataset.update_source(LuxonisSource("test"))
        assert (
            dataset.source.to_document() == LuxonisSource("test").to_document()
        )

    with subtests.test("test_load", bucket_storage=bucket_storage):
        loader = LuxonisLoader(dataset)
        for img, labels in loader:
            assert img is not None
            for task in TASKS:
                assert task in labels

    with subtests.test("test_load_aug", bucket_storage=bucket_storage):
        aug_config = [
            {
                "name": "Mosaic4",
                "params": {"out_width": 416, "out_height": 416, "p": 1.0},
            },
            {"name": "Defocus", "params": {"p": 1.0}},
            {"name": "Sharpen", "params": {"p": 1.0}},
            {"name": "Flip", "params": {"p": 1.0}},
            {"name": "RandomRotate90", "params": {"p": 1.0}},
        ]
        augmentations = Augmentations([512, 512], aug_config)
        loader = LuxonisLoader(dataset, augmentations=augmentations)
        for img, labels in loader:
            assert img is not None
            for task in TASKS:
                assert task in labels

    with subtests.test("test_delete", bucket_storage=bucket_storage):
        dataset.delete_dataset(delete_remote=True)
        assert not LuxonisDataset.exists(
            dataset_name, bucket_storage=bucket_storage
        )


@pytest.mark.dependency(name="test_dataset[BucketStorage.LOCAL]")
def test_dataset_fail():
    dataset = LuxonisDataset("__test_fail", delete_existing=True)

    def generator():
        for i in range(10):
            img = make_image(i)
            yield {
                "file": img,
                "annotation": {
                    "type": "classification",
                    "class": "person",
                },
            }

    dataset.add(generator(), batch_size=2)

    with pytest.raises(ValueError):
        dataset.set_skeletons()


@pytest.mark.dependency(name="test_dataset[BucketStorage.LOCAL]")
def test_loader_iterator():
    dataset = LuxonisParser(
        f"{URL_PREFIX}/COCO_people_subset.zip",
        dataset_name="_iterator_test",
        save_dir=WORK_DIR,
        dataset_type=DatasetType.COCO,
        delete_existing=True,
    ).parse()
    loader = LuxonisLoader(dataset)

    def _raise(*_):
        raise IndexError

    loader._load_image_with_annotations = _raise
    with pytest.raises(IndexError):
        _ = loader[0]


@pytest.mark.parametrize(
    ("bucket_storage",),
    [
        (BucketStorage.LOCAL,),
        (BucketStorage.GCS,),
    ],
)
def test_make_splits(
    bucket_storage: BucketStorage, platform_name: str, python_version: str
):
    definitions: Dict[str, List[str]] = defaultdict(list)

    _start_index: int = 0

    def generator(step=15):
        nonlocal _start_index
        definitions.clear()
        for i in range(_start_index, _start_index + step):
            path = make_image(i)
            yield {
                "file": str(path),
                "annotation": {
                    "type": "classification",
                    "class": ["dog", "cat"][i % 2],
                },
            }
            definitions[["train", "val", "test"][i % 3]].append(str(path))
        _start_index += step

    dataset = LuxonisDataset(
        f"_test_split-{bucket_storage.value}-{platform_name}-{python_version}",
        delete_existing=True,
        delete_remote=True,
        bucket_storage=bucket_storage,
    )
    dataset.add(generator())
    assert dataset.get_splits() is None
    dataset.make_splits(definitions)
    splits = dataset.get_splits()
    assert splits is not None
    assert set(splits.keys()) == {"train", "val", "test"}
    for split, split_data in splits.items():
        assert (
            len(split_data) == 5
        ), f"Split {split} has {len(split_data)} samples"

    dataset.add(generator())
    splits = dataset.get_splits()
    assert splits is not None
    for split, split_data in splits.items():
        assert (
            len(split_data) == 5
        ), f"Split {split} has {len(split_data)} samples"
    dataset.make_splits(definitions)
    splits = dataset.get_splits()
    assert splits is not None
    for split, split_data in splits.items():
        assert (
            len(split_data) == 10
        ), f"Split {split} has {len(split_data)} samples"

    dataset.add(generator())
    dataset.make_splits((1, 0, 0))
    splits = dataset.get_splits()
    assert splits is not None
    for split, split_data in splits.items():
        expected_length = 25 if split == "train" else 10
        assert (
            len(split_data) == expected_length
        ), f"Split {split} has {len(split_data)} samples"

    with pytest.raises(ValueError):
        dataset.make_splits()

    with pytest.raises(ValueError):
        dataset.make_splits((0.7, 0.1, 1))

    with pytest.raises(ValueError):
        dataset.make_splits((0.7, 0.1, 0.1, 0.1))  # type: ignore

    with pytest.raises(ValueError):
        dataset.make_splits((0.7, 0.1, 1), definitions=definitions)

    with pytest.raises(ValueError):
        dataset.make_splits({"train": 1.5})

    with pytest.raises(ValueError):
        dataset.make_splits(
            {split: defs * 2 for split, defs in splits.items()}
        )

    dataset.add(generator(10))
    dataset.make_splits({"custom_split": 1.0})
    splits = dataset.get_splits()
    assert splits is not None
    assert set(splits.keys()) == {"train", "val", "test", "custom_split"}
    for split, split_data in splits.items():
        expected_length = 25 if split == "train" else 10
        assert (
            len(split_data) == expected_length
        ), f"Split {split} has {len(split_data)} samples"

    dataset.make_splits(replace_old_splits=True)
    splits = dataset.get_splits()
    assert splits is not None
    for split, split_data in splits.items():
        expected_length = {"train": 44, "val": 6, "test": 5}
        assert (
            len(split_data) == expected_length[split]
        ), f"Split {split} has {len(split_data)} samples"


def test_complex_dataset():
    url = f"{URL_PREFIX}/D1_ParkingSlotTest"
    base_path = LuxonisFileSystem.download(url, WORK_DIR)
    mask_brand_path = base_path / "mask_brand"
    mask_color_path = base_path / "mask_color"
    kpt_mask_path = base_path / "keypoints_mask_vehicle"

    def generator():
        filenames: Dict[int, Path] = {}
        for base_path in [kpt_mask_path, mask_brand_path, mask_color_path]:
            for sequence_path in list(sorted(base_path.glob("sequence.*"))):
                frame_data = sequence_path / "step0.frame_data.json"
                with open(frame_data) as f:
                    data = json.load(f)["captures"][0]
                    frame_data = data["annotations"]
                    sequence_num = int(sequence_path.suffix[1:])
                    filename = data["filename"]
                    if filename is not None:
                        filename = sequence_path / filename
                        filenames[sequence_num] = filename
                    else:
                        filename = filenames[sequence_num]
                    W, H = data["dimension"]

                annotations = {
                    anno["@type"].split(".")[-1]: anno for anno in frame_data
                }

                bbox_classes = {}
                bboxes = {}

                for bbox_annotation in annotations.get(
                    "BoundingBox2DAnnotation", defaultdict(list)
                )["values"]:
                    class_ = (
                        bbox_annotation["labelName"].split("-")[-1].lower()
                    )
                    if class_ == "motorbiek":
                        class_ = "motorbike"
                    x, y = bbox_annotation["origin"]
                    w, h = bbox_annotation["dimension"]
                    instance_id = bbox_annotation["instanceId"]
                    bbox_classes[instance_id] = class_
                    bboxes[instance_id] = [x / W, y / H, w / W, h / H]
                    yield {
                        "file": filename,
                        "annotation": {
                            "type": "boundingbox",
                            "class": class_,
                            "x": x / W,
                            "y": y / H,
                            "w": w / W,
                            "h": h / H,
                            "instance_id": instance_id,
                        },
                    }

                for kpt_annotation in annotations.get(
                    "KeypointAnnotation", defaultdict(list)
                )["values"]:
                    keypoints = kpt_annotation["keypoints"]
                    instance_id = kpt_annotation["instanceId"]
                    class_ = bbox_classes[instance_id]
                    bbox = bboxes[instance_id]
                    kpts = []

                    if class_ == "motorbike":
                        keypoints = keypoints[:3]
                    else:
                        keypoints = keypoints[3:]

                    for kp in keypoints:
                        x, y = kp["location"]
                        kpts.append([x / W, y / H, kp["state"]])

                    yield {
                        "file": filename,
                        "annotation": {
                            "type": "detection",
                            "class": class_,
                            "task": class_,
                            "keypoints": kpts,
                            "instance_id": instance_id,
                            "boundingbox": {
                                "x": bbox[0],
                                "y": bbox[1],
                                "w": bbox[2],
                                "h": bbox[3],
                            },
                        },
                    }

                vehicle_type_segmentation = annotations[
                    "SemanticSegmentationAnnotation"
                ]
                mask = cv2.cvtColor(
                    cv2.imread(
                        str(
                            sequence_path
                            / vehicle_type_segmentation["filename"]
                        )
                    ),
                    cv2.COLOR_BGR2RGB,
                )
                classes = {
                    inst["labelName"]: inst["pixelValue"][:3]
                    for inst in vehicle_type_segmentation["instances"]
                }
                if base_path == kpt_mask_path:
                    task = "vehicle_type_segmentation"
                elif base_path == mask_brand_path:
                    task = "brand_segmentation"
                else:
                    task = "color_segmentation"
                for class_, mask_ in rgb_to_bool_masks(
                    mask, classes, add_background_class=True
                ):
                    yield {
                        "file": filename,
                        "annotation": {
                            "type": "mask",
                            "class": class_,
                            "task": task,
                            "mask": mask_,
                        },
                    }
                if base_path == mask_color_path:
                    yield {
                        "file": filename,
                        "annotation": {
                            "type": "mask",
                            "class": "vehicle",
                            "task": "vehicle_segmentation",
                            "mask": mask.astype(bool)[..., 0]
                            | mask.astype(bool)[..., 1]
                            | mask.astype(bool)[..., 2],
                        },
                    }

    dataset = LuxonisDataset("__D1ParkingSLot-test", delete_existing=True)
    dataset.add(generator())
    dataset.make_splits()
    assert len(dataset) == 200
    assert set(dataset.get_tasks()) == {
        "boundingbox",
        "motorbike-boundingbox",
        "motorbike-keypoints",
        "car-boundingbox",
        "car-keypoints",
        "vehicle_type_segmentation",
        "brand_segmentation",
        "color_segmentation",
        "vehicle_segmentation",
    }
    assert dataset.get_classes()[1] == {
        "boundingbox": sorted(["motorbike", "car"]),
        "motorbike-boundingbox": ["motorbike"],
        "motorbike-keypoints": ["motorbike"],
        "car-boundingbox": ["car"],
        "car-keypoints": ["car"],
        "vehicle_type_segmentation": ["background", "car", "motorbike"],
        "brand_segmentation": sorted(
            [
                "background",
                "chrysler",
                "bmw",
                "ducati",
                "dodge",
                "ferrari",
                "infiniti",
                "land-rover",
                "roll-royce",
                "saab",
                "Kawasaki",
                "moto",
                "truimph",
                "alfa-romeo",
                "harley",
                "honda",
                "jeep",
                "aprilia",
                "piaggio",
                "yamaha",
                "buick",
                "pontiac",
                "isuzu",
            ]
        ),
        "color_segmentation": ["background", "blue", "green", "red"],
        "vehicle_segmentation": ["vehicle"],
    }


@pytest.mark.parametrize(
    ("bucket_storage",),
    [
        (BucketStorage.LOCAL,),
        (BucketStorage.GCS,),
    ],
)
def test_uncommon_label_types(
    bucket_storage: BucketStorage, platform_name: str, python_version: str
):
    arr = np.random.rand(10)

    def generator():
        for i in range(10):
            img = make_image(i)
            yield {
                "file": img,
                "annotation": {
                    "type": "label",
                    "value": "dog",
                },
            }
            yield {
                "file": img,
                "annotation": {
                    "type": "label",
                    "value": "cat",
                },
            }
            np.save(str(img.with_suffix(".npy")), arr)
            yield {
                "file": img,
                "annotation": {
                    "type": "array",
                    "path": str(img.with_suffix(".npy")),
                },
            }

    dataset_name = f"__uncommon_label_types-{bucket_storage.value}-{platform_name}-{python_version}"
    dataset = LuxonisDataset(
        dataset_name,
        delete_existing=True,
        delete_remote=True,
        bucket_storage=bucket_storage,
    )
    dataset.add(generator())
    dataset.make_splits()
    loader = LuxonisLoader(dataset)
    for _, labels in loader:
        assert "label" in labels
        assert "array" in labels
        assert labels["label"][0].tolist() == [["dog"], ["cat"]]
        assert np.allclose(labels["array"][0], arr)


def test_no_labels():
    dataset = LuxonisDataset("__no_labels", delete_existing=True)

    def generator():
        for i in range(10):
            img = make_image(i)
            yield {
                "file": img,
            }

    dataset.add(generator())
    dataset.make_splits()

    loader = LuxonisLoader(dataset)
    for _, labels in loader:
        assert labels == {}

    augments = Augmentations([512, 512], [{"name": "Flip", "params": {}}])
    loader = LuxonisLoader(dataset, augmentations=augments)
    for _, labels in loader:
        assert labels == {}



from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

from PIL import Image

from luxonis_ml.data import DatasetIterator

from .base_parser import BaseParser, ParserOutput


class YoloV4Parser(BaseParser):
    """Parses directory with YoloV4 annotations to LDF.

    Expected format::

        dataset_dir/
        ├── train/
        │   ├── img1.jpg
        │   ├── img2.jpg
        │   ├── ...
        │   ├── _annotations.txt
        │   └── _classes.txt
        ├── valid/
        └── test/

    This is one of the formats that can be generated by
    U{Roboflow <https://roboflow.com/>}.
    """

    @staticmethod
    def validate_split(split_path: Path) -> Optional[Dict[str, Any]]:
        if not split_path.exists():
            return None
        annotations = split_path / "_annotations.txt"
        classes = split_path / "_classes.txt"
        if not annotations.exists() or not classes.exists():
            return None
        return {
            "image_dir": split_path,
            "annotation_path": annotations,
            "classes_path": classes,
        }

    @staticmethod
    def validate(dataset_dir: Path) -> bool:
        for split in ["train", "valid", "test"]:
            split_path = dataset_dir / split
            if YoloV4Parser.validate_split(split_path) is None:
                return False
        return True

    def from_dir(
        self, dataset_dir: Path
    ) -> Tuple[List[str], List[str], List[str]]:
        added_train_imgs = self._parse_split(
            image_dir=dataset_dir / "train",
            annotation_path=dataset_dir / "train" / "_annotations.txt",
            classes_path=dataset_dir / "train" / "_classes.txt",
        )
        added_val_imgs = self._parse_split(
            image_dir=dataset_dir / "valid",
            annotation_path=dataset_dir / "valid" / "_annotations.txt",
            classes_path=dataset_dir / "valid" / "_classes.txt",
        )
        added_test_imgs = self._parse_split(
            image_dir=dataset_dir / "test",
            annotation_path=dataset_dir / "test" / "_annotations.txt",
            classes_path=dataset_dir / "test" / "_classes.txt",
        )
        return added_train_imgs, added_val_imgs, added_test_imgs

    def from_split(
        self, image_dir: Path, annotation_path: Path, classes_path: Path
    ) -> ParserOutput:
        """Parses annotations from YoloV4 format to LDF. Annotations
        include classification and object detection.

        @type image_dir: Path
        @param image_dir: Path to directory with images
        @type annotation_path: Path
        @param annotation_path: Path to annotation file
        @type classes_path: Path
        @param classes_path: Path to file with class names
        @rtype: L{ParserOutput}
        @return: Annotation generator, list of classes names, skeleton
            dictionary for keypoints and list of added images.
        """
        with open(classes_path) as f:
            class_names = {
                i: line.rstrip() for i, line in enumerate(f.readlines())
            }

        def generator() -> DatasetIterator:
            with open(annotation_path) as f:
                annotation_data = [line.rstrip() for line in f.readlines()]

            for ann_line in annotation_data:
                data = ann_line.split(" ")
                img_path = data[0]

                path = image_dir.absolute().resolve() / img_path
                if not path.exists():
                    continue

                file = str(path)

                img = Image.open(file)
                width, height = img.size

                for ann_data in data[1:]:
                    curr_ann_data = ann_data.split(",")
                    class_name = class_names[int(curr_ann_data[4])]
                    yield {
                        "file": file,
                        "annotation": {
                            "type": "classification",
                            "class": class_name,
                        },
                    }

                    bbox_xyxy = [float(i) for i in curr_ann_data[:4]]
                    yield {
                        "file": file,
                        "annotation": {
                            "type": "boundingbox",
                            "class": class_name,
                            "x": bbox_xyxy[0] / width,
                            "y": bbox_xyxy[1] / height,
                            "w": (bbox_xyxy[2] - bbox_xyxy[0]) / width,
                            "h": (bbox_xyxy[3] - bbox_xyxy[1]) / height,
                        },
                    }

        added_images = self._get_added_images(generator())

        return generator(), list(class_names.values()), {}, added_images



import random
from typing import Any, Dict, List, Optional, Union, cast

import numpy as np
from albumentations.core.bbox_utils import (
    BboxParams,
    BboxProcessor,
    DataProcessor,
)
from albumentations.core.composition import (
    BaseCompose,
    TransformsSeqType,
    get_always_apply,
)
from albumentations.core.keypoints_utils import (
    KeypointParams,
    KeypointsProcessor,
)
from albumentations.core.utils import get_shape

from .batch_processors import BboxBatchProcessor, KeypointsBatchProcessor
from .batch_utils import (
    batch2list,
    concat_batches,
    list2batch,
    to_unbatched_name,
)


class Compose(BaseCompose):
    def __init__(
        self,
        transforms: TransformsSeqType,
        bbox_params: Optional[Union[dict, BboxParams]] = None,
        keypoint_params: Optional[Union[dict, KeypointParams]] = None,
        additional_targets: Optional[Dict[str, str]] = None,
        p: float = 1.0,
        is_check_shapes: bool = True,
    ):
        """Compose transforms and handle all transformations regarding
        bounding boxes.

        @param transforms: List of transformations to compose
        @type transforms: TransformsSeqType
        @param bbox_params: Parameters for bounding boxes transforms. Defaults to None.
        @type bbox_params: Optional[Union[dict, BboxParams]]
        @param keypoint_params: Parameters for keypoint transforms. Defaults to None.
        @type keypoint_params: Optional[Union[dict, KeypointParams]]
        @param additional_targets: Dict with keys - new target name, values - old target
        name. ex: {'image2': 'image'}. Defaults to None.
        @type additional_targets: Optional[Dict[str, str]]
        @param p: Probability of applying all list of transforms. Defaults to 1.0.
        @type p: float
        @param is_check_shapes: If True shapes consistency of images/mask/masks would be checked on
        each call. If you would like to disable this check - pass False (do it only if you are sure
        in your data consistency). Defaults to True.
        @type is_check_shapes: bool
        """
        super().__init__(transforms, p)

        self.processors: Dict[str, DataProcessor] = {}

        if bbox_params:
            if isinstance(bbox_params, dict):
                b_params = BboxParams(**bbox_params)
            elif isinstance(bbox_params, BboxParams):
                b_params = bbox_params
            else:
                raise ValueError(
                    "unknown format of bbox_params, please use `dict` or `BboxParams`"
                )
            self.processors["bboxes"] = self._get_bbox_processor(
                b_params, additional_targets
            )

        if keypoint_params:
            if isinstance(keypoint_params, dict):
                k_params = KeypointParams(**keypoint_params)
            elif isinstance(keypoint_params, KeypointParams):
                k_params = keypoint_params
            else:
                raise ValueError(
                    "unknown format of keypoint_params, please use `dict` or `KeypointParams`"
                )
            self.processors["keypoints"] = self._get_keypoints_processor(
                k_params, additional_targets
            )

        if additional_targets is None:
            additional_targets = {}

        self.additional_targets = additional_targets

        for proc in self.processors.values():
            proc.ensure_transforms_valid(self.transforms)

        self.add_targets(additional_targets)

        self.is_check_args = True
        self._disable_check_args_for_transforms(self.transforms)

        self.is_check_shapes = is_check_shapes

    def _get_bbox_processor(self, b_params, additional_targets):
        return BboxProcessor(b_params, additional_targets)

    def _get_keypoints_processor(self, k_params, additional_targets):
        return KeypointsProcessor(k_params, additional_targets)

    @staticmethod
    def _disable_check_args_for_transforms(
        transforms: TransformsSeqType,
    ) -> None:
        for transform in transforms:
            if isinstance(transform, BaseCompose):
                Compose._disable_check_args_for_transforms(
                    transform.transforms
                )
            if isinstance(transform, Compose):
                transform._disable_check_args()

    def _disable_check_args(self) -> None:
        self.is_check_args = False

    def __call__(
        self, *args, force_apply: bool = False, **data
    ) -> Dict[str, Any]:
        if args:
            raise KeyError(
                "You have to pass data to augmentations as named arguments, for example: aug(image=image)"
            )
        if self.is_check_args:
            self._check_args(self.additional_targets, **data)
        assert isinstance(
            force_apply, (bool, int)
        ), "force_apply must have bool or int type"
        need_to_run = force_apply or random.random() < self.p
        for p in self.processors.values():
            p.ensure_data_valid(data)
        transforms = (
            self.transforms
            if need_to_run
            else get_always_apply(self.transforms)
        )

        check_each_transform = any(
            getattr(item.params, "check_each_transform", False)
            for item in self.processors.values()
        )

        for p in self.processors.values():
            p.preprocess(data)

        for _, t in enumerate(transforms):
            data = t(**data)
            if check_each_transform:
                data = self._check_data_post_transform(data)

        data = self._make_targets_contiguous(
            data
        )  # ensure output targets are contiguous

        for p in self.processors.values():
            p.postprocess(data)

        return data

    def _check_data_post_transform(
        self, data: Dict[str, Any]
    ) -> Dict[str, Any]:
        rows, cols = get_shape(data["image"])

        for p in self.processors.values():
            if not getattr(p.params, "check_each_transform", False):
                continue

            for data_name in p.data_fields:
                data[data_name] = p.filter(data[data_name], rows, cols)
        return data

    def _to_dict(self) -> Dict[str, Any]:
        dictionary = super(Compose, self)._to_dict()
        bbox_processor = self.processors.get("bboxes")
        keypoints_processor = self.processors.get("keypoints")
        dictionary.update(
            {
                "bbox_params": bbox_processor.params._to_dict()
                if bbox_processor
                else None,  # skipcq: PYL-W0212
                "keypoint_params": keypoints_processor.params._to_dict()  # skipcq: PYL-W0212
                if keypoints_processor
                else None,
                "additional_targets": self.additional_targets,
                "is_check_shapes": self.is_check_shapes,
            }
        )
        return dictionary

    def get_dict_with_id(self) -> Dict[str, Any]:
        dictionary = super().get_dict_with_id()
        bbox_processor = self.processors.get("bboxes")
        keypoints_processor = self.processors.get("keypoints")
        dictionary.update(
            {
                "bbox_params": bbox_processor.params._to_dict()
                if bbox_processor
                else None,  # skipcq: PYL-W0212
                "keypoint_params": keypoints_processor.params._to_dict()  # skipcq: PYL-W0212
                if keypoints_processor
                else None,
                "additional_targets": self.additional_targets,
                "params": None,
                "is_check_shapes": self.is_check_shapes,
            }
        )
        return dictionary

    def _check_args(self, additional_targets, **kwargs) -> None:
        checked_single = ["image", "mask"]
        checked_multi = ["masks"]
        check_bbox_param = ["bboxes"]
        # ["bboxes", "keypoints"] could be almost any type, no need to check them
        shapes = []
        for data_name, data in kwargs.items():
            internal_data_name = additional_targets.get(data_name, data_name)
            if internal_data_name in checked_single:
                if not isinstance(data, np.ndarray):
                    raise TypeError(
                        "{} must be numpy array type".format(data_name)
                    )
                shapes.append(data.shape[:2])
            if internal_data_name in checked_multi:
                if data is not None:
                    if not isinstance(data[0], np.ndarray):
                        raise TypeError(
                            "{} must be list of numpy arrays".format(data_name)
                        )
                    shapes.append(data[0].shape[:2])
            if (
                internal_data_name in check_bbox_param
                and self.processors.get("bboxes") is None
            ):
                raise ValueError(
                    "bbox_params must be specified for bbox transformations"
                )

        if (
            self.is_check_shapes
            and shapes
            and shapes.count(shapes[0]) != len(shapes)
        ):
            raise ValueError(
                "Height and Width of image, mask or masks should be equal. You can disable shapes check "
                "by setting a parameter is_check_shapes=False of Compose class (do it only if you are sure "
                "about your data consistency)."
            )

    def _make_targets_contiguous(self, data: Dict[str, Any]) -> Dict[str, Any]:
        result = {}
        for key, value in data.items():
            if isinstance(value, np.ndarray):
                value = np.ascontiguousarray(value)
            result[key] = value
        return result


class BatchCompose(Compose):
    def __init__(
        self,
        transforms: TransformsSeqType,
        bbox_params: Optional[Union[dict, BboxParams]] = None,
        keypoint_params: Optional[Union[dict, KeypointParams]] = None,
        additional_targets: Optional[Dict[str, str]] = None,
        p: float = 1.0,
        is_check_shapes: bool = True,
    ):
        """Compose designed to handle the multi-image transforms The
        contents can be a subclass of `BatchBasedTransform` or other
        transforms enclosed by ForEach container. All targets' names
        should have the suffix "_batch", ex ("image_batch",
        "bboxes_batch"). Note this nameing rule is applied to the
        `label_fields` of the `BboxParams` and the `KeypointsParams`.

        @param transforms: List of transformations to compose
        @type transforms: TransformsSeqType
        @param bbox_params: Parameters for bounding boxes transforms. Defaults to None.
        @type bbox_params: Optional[Union[dict, BboxParams]]
        @param keypoint_params: Parameters for keypoint transforms. Defaults to None.
        @type keypoint_params: Optional[Union[dict, KeypointParams]]
        @param additional_targets: Dict with keys - new target name, values - old target
        name. ex: {'image2': 'image'}. Defaults to None.
        @type additional_targets: Optional[Dict[str, str]]
        @param p: Probability of applying all list of transforms. Defaults to 1.0.
        @type p: float
        @param is_check_shapes: If True shapes consistency of images/mask/masks would be checked on
        each call. If you would like to disable this check - pass False (do it only if you are sure
        in your data consistency). Defaults to True.
        @type is_check_shapes: bool
        """
        super().__init__(
            transforms=transforms,
            bbox_params=bbox_params,
            keypoint_params=keypoint_params,
            additional_targets=additional_targets,
            p=p,
            is_check_shapes=is_check_shapes,
        )

    def _get_bbox_processor(self, b_params, additional_targets):
        return BboxBatchProcessor(b_params, additional_targets)

    def _get_keypoints_processor(self, k_params, additional_targets):
        return KeypointsBatchProcessor(k_params, additional_targets)

    def _check_data_post_transform(
        self, batched_data: Dict[str, Any]
    ) -> Dict[str, Any]:
        datalist = batch2list(batched_data)
        processed = []
        for data in datalist:
            rows, cols = get_shape(data["image"])
            for p in self.processors.values():
                if not getattr(p.params, "check_each_transform", False):
                    continue
                p = cast(Union[BboxBatchProcessor, KeypointsBatchProcessor], p)
                for data_name in p.item_processor.data_fields:
                    data[data_name] = p.filter(data[data_name], rows, cols)
            processed.append(data)
        return list2batch(processed)

    def _check_args(self, additional_targets, **kwargs) -> None:
        datalist = batch2list(kwargs)
        unbatched_targets = {
            to_unbatched_name(k): to_unbatched_name(v)
            for k, v in additional_targets.items()
        }
        for data in datalist:
            super()._check_args(unbatched_targets, **data)

    def _make_targets_contiguous(
        self, batched_data: Dict[str, Any]
    ) -> Dict[str, Any]:
        datalist = batch2list(batched_data)
        if len(datalist) == 0:
            return batched_data
        processed = []
        for data in datalist:
            data = super()._make_targets_contiguous(data)
            processed.append(data)
        return list2batch(processed)


class ForEach(BaseCompose):
    """Apply transforms for each batch element This expects batched
    input and can be contained by the `BatchCompose`."""

    def __init__(self, transforms: TransformsSeqType, p: float = 0.5):
        super().__init__(transforms, p)

    def __call__(
        self, *args, force_apply: bool = False, **batched_data
    ) -> Dict[str, List]:
        datalist = batch2list(batched_data)
        processed = []
        for data in datalist:
            for t in self.transforms:
                data = t(force_apply=force_apply, **data)
            processed.append(data)
        batched_data = list2batch(processed)
        return batched_data

    def add_targets(
        self, additional_targets: Optional[Dict[str, str]]
    ) -> None:
        if additional_targets:
            unbatched_targets = {
                to_unbatched_name(k): to_unbatched_name(v)
                for k, v in additional_targets.items()
            }
            for t in self.transforms:
                t.add_targets(unbatched_targets)


class Repeat(BaseCompose):
    """Apply transforms repeatedly and concatenates the output batches.

    This expects batched input and can be contained by the `BatchCompose`.
    The contained transforms should be a subbclass of the `BatchBasedTransform`.
    Internally, this container works as the following way:
    Note: This class assumes that each transform does not modify the input data.
    """

    def __init__(self, transforms: TransformsSeqType, n: int, p: float = 0.5):
        super().__init__(transforms, p)
        if n <= 0:
            raise ValueError("Repetition `n` should be larger than 0")
        self.n = n

    def __call__(
        self, *args, force_apply: bool = False, **batched_data
    ) -> Dict[str, List]:
        processed = []
        for _ in range(self.n):
            image = batched_data["image_batch"][0].copy()
            data = batched_data
            for t in self.transforms:
                data = t(force_apply=force_apply, **data)
            processed.append(data)
            assert np.all(batched_data["image_batch"][0] == image)
        return concat_batches(processed)



"""Near-duplicate Search with Qdrant and Weaviate.

Overview:
This module provides utilities to detect and remove near-duplicate data points within a given set of embeddings.
It leverages vector databases (Qdrant or Weaviate) for efficient search and retrieval,
and employs Kernel Density Estimation (KDE) for optimal split based on embeddings' cosine similarity.
This approach is particularly well-suited for handling high-dimensional embeddings.

Key Features:
    - Vector Database Integration: Supports both Qdrant and Weaviate for flexible deployment options.
    - KDE-Based Near-Duplicate Detection: Uses KDE to identify clusters of near-duplicates, ensuring accuracy in high-dimensional spaces.
    - Visualization: Allows plotting KDE results using matplotlib for intuitive understanding.
    - Dynamic KDE Peak Selection: Automatically determines the best candidates for removal based on KDE peaks, minimizing manual thresholding.

Dependencies:
    - KDEpy
    - Qdrant (optional, for Qdrant-specific features)
    - Weaviate (optional, for Weaviate-specific features)

Functions:
    - search_vectordb(vectordb_api, query_vector, property_name, top_k): Searches for similar embeddings within the specified vector database.
    - _plot_kde(xs, s, density, maxima, minima): Plots a KDE distribution.
    - kde_peaks(data, bandwidth="scott", plot=False): Identifies peaks in a KDE distribution.
    - find_similar(reference_embeddings, vectordb_api, k=100, n=1000, method="first", k_method=None, kde_bw="scott", plot=False): Finds the most similar embeddings to the given reference embeddings.

Examples (using Qdrant):
    1. Initialize a Qdrant client and retrieve embeddings:
        from luxonis_ml.embeddings.utils.qdrant import QdrantAPI
        qdrant_api = QdrantAPI(host="localhost", port=6333)
        qdrant_api.create_collection("images", ["image_path", "embedding"])
        id_X, X = qdrant_api.get_all_embeddings()

    2. Find similar embeddings using various methods:
        # By instance ID:
        ix, paths = find_similar_qdrant(id_X[i], qdrant_api, "image_path", 5, 100, "first")

        # Using KDE Peaks method:
        ix, paths = find_similar_qdrant(X[i], qdrant_api, "image_path", 5, 100, "first", "kde_peaks", "silverman", plot=False)

        # Based on average of multiple embeddings:
        dark_ix = np.array([10, 123, 333, 405])
        emb_dark = X[dark_ix]
        remove_dark_ix, paths = find_similar_qdrant(emb_dark, qdrant_api, "image_path", 25, 5000, "average", "kde_basic", "scott", plot=True)

Additional Notes:
    - For Weaviate-specific examples, refer to the provided code examples.
    - The search_vectordb function can be used with either Qdrant or Weaviate, depending on the provided vectordb_api object.
    - Adjust parameters like k, n, and kde_bw based on your dataset and requirements.
"""

from typing import List, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from KDEpy import FFTKDE
from scipy.signal import argrelextrema

from luxonis_ml.embeddings.utils.vectordb import VectorDBAPI


def _plot_kde(
    xs: np.ndarray,
    s: np.ndarray,
    density: np.ndarray,
    maxima: np.ndarray,
    minima: np.ndarray,
) -> None:
    """Plot a KDE distribution.

    @type xs: np.ndarray
    @param xs: The x-axis values.
    @type s: np.ndarray
    @param s: The y-axis values.
    @type density: np.ndarray
    @param density: The density values.
    @type maxima: np.ndarray
    @param maxima: The indices of the local maxima.
    @type minima: np.ndarray
    @param minima: The indices of the local minima.
    """
    plt.plot(xs, density, label="KDE")
    plt.plot(xs[maxima], s[maxima], "ro", label="local maxima")
    plt.plot(xs[minima], s[minima], "bo", label="local minima")
    plt.plot(xs[np.argmax(s)], np.max(s), "go", label="global maxima")
    plt.legend()
    plt.show()


def kde_peaks(
    data: np.ndarray,
    bandwidth: Union[str, float] = "scott",
    plot: bool = False,
) -> Tuple[np.ndarray, np.ndarray, int, float]:
    """Find peaks in a KDE distribution using scipy's argrelextrema
    function.

    @type data: np.ndarray
    @param data: The data to fit the KDE.
    @type bandwidth: Union[str, float]
    @param bandwidth: The bandwidth to use for the KDE. Default is
        'scott'.
    @type plot: bool
    @param plot: Whether to plot the KDE.
    @rtype: Tuple[np.ndarray, np.ndarray, int, float]
    @return: The indices of the KDE maxima, the indices of the KDE
        minima, the index of the global maxima, and the standard
        deviation of the data.
    """
    # fit density
    kde = FFTKDE(kernel="gaussian", bw=bandwidth)
    xs = np.linspace(np.min(data) - 0.01, np.max(data) + 0.01, 1000)
    density = kde.fit(data).evaluate(xs)

    # find local maxima
    s = np.array(density)
    maxima = argrelextrema(s, np.greater)
    minima = argrelextrema(s, np.less)

    # of the local maxima, find the global maxima index
    global_max_ix = np.argmax(s[maxima])

    # find variance of distribution
    std = np.std(data)

    # plot
    if plot:
        _plot_kde(xs, s, density, maxima, minima)

    return xs[maxima], xs[minima], global_max_ix, std


def find_similar(
    reference_embeddings: Union[str, List[str], List[List[float]], np.ndarray],
    vectordb_api: VectorDBAPI,
    k: int = 100,
    n: int = 1000,
    method: str = "first",
    k_method: Union[str, None] = None,
    kde_bw: Union[str, float] = "scott",
    plot: bool = False,
) -> np.ndarray:
    """Find the most similar embeddings to the reference embeddings.

    @type reference_embeddings: Union[str, List[str], List[List[float]],
        np.ndarray]
    @param reference_embeddings: The embeddings to compare against. Or a
        list of of embedding instance_ids that reside in VectorDB.
    @type vectordb_api: VectorDBAPI
    @param vectordb_api: The VectorDBAPI instance to use.
    @type k: int
    @param k: The number of embeddings to return. Default is 100.
    @type n: int
    @param n: The number of embeddings to compare against. Default is
        1000. (This is the number of embeddings that are returned by the
        VectorDB search. It matters for the KDE, as it can be slow for
        large n. Your choice of n depends on the amount of duplicates in
        your dataset, the more duplicates, the larger n should be. If
        you have 2-10 duplicates per image, n=100 should be ok. If you
        have 50-300 duplicates per image, n=1000 should work good
        enough.
    @type method: str
    @param method: The method to use to find the most similar
        embeddings. If 'first' use the first of the reference
        embeddings. If 'average', use the average of the reference
        embeddings.
    @type k_method: str
    @param k_method: The method to select the best k. If None, use k as
        is. If 'kde_basic', use the minimum of the KDE. If 'kde_peaks',
        use the minimum of the KDE peaks, according to a specific
        hardcoded hevristics/thresholds.
    @type kde_bw: Union[str, float]
    @param kde_bw: The bandwidth to use for the KDE. Default is 'scott'.
    @type plot: bool
    @param plot: Whether to plot the KDE.
    @rtype: np.array
    @return: The instance_ids of the most similar embeddings.
    """

    # Get the reference embeddings
    if isinstance(
        reference_embeddings, str
    ):  # if it is a single instance_id: string uuid
        reference_embeddings = [reference_embeddings]
    if isinstance(
        reference_embeddings[0], str
    ):  # if it is a list of instance_ids: list of strings
        reference_embeddings = vectordb_api.retrieve_embeddings_by_ids(
            reference_embeddings
        )

    # Select the reference embedding
    if method == "first":
        if isinstance(
            reference_embeddings, list
        ):  # if it is a list of embeddings: list of lists of floats
            reference_embeddings = np.array(reference_embeddings)
        if len(reference_embeddings.shape) > 1:
            reference_embeddings = reference_embeddings[0]
    elif method == "average":
        # Calculate the average of the reference embeddings
        avg_embedding = np.mean(reference_embeddings, axis=0)
        reference_embeddings = avg_embedding
    else:
        raise ValueError(f"Unknown method: {method}")

    ix, similarities = vectordb_api.search_similar_embeddings(
        reference_embeddings, top_k=n
    )
    ix, similarities = np.array(ix), np.array(similarities)

    # Select the best k embeddings
    if k_method is None:
        best_embeddings_ix = np.argsort(similarities)[-k:]

    elif k_method == "kde_basic":
        _, new_min, _, _ = kde_peaks(similarities, bandwidth=kde_bw, plot=plot)

        if len(new_min) > 0:
            k = len(np.where(similarities > new_min[-1])[0])
            if k < 2 and len(new_min) > 1:
                k = len(np.where(similarities > new_min[-2])[0])
            print(k, new_min)

        best_embeddings_ix = np.argsort(similarities)[-k:]

    elif k_method == "kde_peaks":
        if len(similarities) > 50000:
            print("Too many embeddings, using 97 percentile")
            # take top 97 procentile of closest points
            p_97 = np.percentile(similarities, 97)
            ix_sim = np.where(similarities > p_97)[0]
            _, new_min, _, _ = kde_peaks(
                similarities[ix_sim], bandwidth=kde_bw, plot=plot
            )

            if len(new_min) > 0:
                k = len(np.where(similarities > new_min[-1])[0])
                print(k, new_min)

        else:
            # get maxima and minima of the KDE on the distances
            _, minima, _, _ = kde_peaks(
                similarities, bandwidth=kde_bw, plot=plot
            )
            if len(minima) > 0:
                minima = minima[-1]
                if minima < 0.94:
                    minima = 0.97
                else:
                    minima = max(minima, 0.96)

            else:
                minima = 0.97

            # select k closest embeddings
            k = len(np.where(similarities > minima)[0])
            if minima > 0.97 and k < 2:
                k = len(np.where(similarities > 0.97)[0])
            print(k, minima)

        # select the best k embeddings
        best_embeddings_ix = np.argsort(similarities)[-k:]

    else:
        raise ValueError(f"Unknown k_method: {k_method}")

    return ix[best_embeddings_ix]
